/*
 *	binkleyforce -- unix FTN mailer project
 *	
 *	Copyright (c) 1998-2000 Alexander Belkin, 2:5020/1398.11
 *	
 *	This program is free software; you can redistribute it and/or modify
 *	it under the terms of the GNU General Public License as published by
 *	the Free Software Foundation; either version 2 of the License, or
 *	(at your option) any later version.
 *	
 *	$Id: u_recode.c,v 1.1.1.1 2001/08/30 13:33:51 kst Exp $
 */

#include "includes.h"
#include "confread.h"
#include "logger.h"
#include "util.h"

char default_table[RECODE_MAX_CHAR+1] =
  "\x00\x01\x02\x03\x04\x05\x06\x07\x08\x09\x0a\x0b\x0c\x0d\x0e\x0f"
  "\x10\x11\x12\x13\x14\x15\x16\x17\x18\x19\x1a\x1b\x1c\x1d\x1e\x1f"
  "\x20\x21\x22\x23\x24\x25\x26\x27\x28\x29\x2a\x2b\x2c\x2d\x2e\x2f"
  "\x30\x31\x32\x33\x34\x35\x36\x37\x38\x39\x3a\x3b\x3c\x3d\x3e\x3f"
  "\x40\x41\x42\x43\x44\x45\x46\x47\x48\x49\x4a\x4b\x4c\x4d\x4e\x4f"
  "\x50\x51\x52\x53\x54\x55\x56\x57\x58\x59\x5a\x5b\x5c\x5d\x5e\x5f"
  "\x60\x61\x62\x63\x64\x65\x66\x67\x68\x69\x6a\x6b\x6c\x6d\x6e\x6f"
  "\x70\x71\x72\x73\x74\x75\x76\x77\x78\x79\x7a\x7b\x7c\x7d\x7e\x7f"
  "\x80\x81\x82\x83\x84\x85\x86\x87\x88\x89\x8a\x8b\x8c\x8d\x8e\x8f"
  "\x90\x91\x92\x93\x94\x95\x96\x97\x98\x99\x9a\x9b\x9c\x9d\x9e\x9f"
  "\xa0\xa1\xa2\xa3\xa4\xa5\xa6\xa7\xa8\xa9\xaa\xab\xac\xad\xae\xaf"
  "\xb0\xb1\xb2\xb3\xb4\xb5\xb6\xb7\xb8\xb9\xba\xbb\xbc\xbd\xbe\xbf"
  "\xc0\xc1\xc2\xc3\xc4\xc5\xc6\xc7\xc8\xc9\xca\xcb\xcc\xcd\xce\xcf"
  "\xd0\xd1\xd2\xd3\xd4\xd5\xd6\xd7\xd8\xd9\xda\xdb\xdc\xdd\xde\xdf"
  "\xe0\xe1\xe2\xe3\xe4\xe5\xe6\xe7\xe8\xe9\xea\xeb\xec\xed\xee\xef"
  "\xf0\xf1\xf2\xf3\xf4\xf5\xf6\xf7\xf8\xf9\xfa\xfb\xfc\xfd\xfe\xff";

s_recode_table recode_table_in;
s_recode_table recode_table_out;
bool initialised = FALSE;

static void recode_init(void)
{
	memset(&recode_table_in, '\0', sizeof(s_recode_table));
	memset(&recode_table_out, '\0', sizeof(s_recode_table));
	
	memcpy(recode_table_in.table, default_table, sizeof(default_table));
	memcpy(recode_table_out.table, default_table, sizeof(default_table));
}

static int recode_load_table(char *dest, const char *filename)
{
	FILE *fp;
	char buf[80];
	int find = 0;
	int repl = 0;
	int line = 0;

	/* Restore default recode table */
	memcpy(dest, default_table, sizeof(default_table));
	
	fp = file_open(filename, "r");
	if( !fp )
	{
		logerr("cannot open recode table \"%s\"", filename);
		return -1;
	}
	
	while( fgets(buf, sizeof(buf), fp) )
	{
		++line;
		string_chomp(buf);
		string_trimboth(buf);

		if( *buf == '#' )
			continue;

		if( sscanf(buf, "%i %i", &find, &repl) == 2 )
		{
			if( 0 > find || find > RECODE_MAX_CHAR )
				log("recode character out of range 0..%d in \"%s\" line %d",
					RECODE_MAX_CHAR, filename, line);
			else
				dest[find] = (unsigned char)repl;
		}
	}

	file_close(fp);

	return 0;
}

static int recode_update_table(s_recode_table *tab, const char *filename)
{
	bool newtab = FALSE;
	
	/*
	 * Check/update current recode table
	 */
	if( filename && *filename )
	{
		if( !tab->filename || strcmp(filename, tab->filename) )
			newtab = TRUE;
	}
	else if( tab->filename && *tab->filename )
		newtab = TRUE;

	if( newtab )
	{
		if( filename && *filename )
		{
			tab->filename = filename;
			recode_load_table(tab->table, filename);
		}
		else
		{
			tab->filename = NULL;
			memcpy(tab->table, default_table, sizeof(default_table));
		}
	}

	return 0;
}

char *recode(char *str, s_recode_table *tab, const char *filename)
{
	char *p;
	
	if( !initialised )
	{
		recode_init();
		initialised = TRUE;
	}
	
	recode_update_table(tab, filename);
	
	for( p = str; *p; p++ )
		*p = tab->table[(unsigned char)*p];
	
	return str;
}

