
/* Do you want debug code to be compiled? */
#undef DEBUG

/* Do you want use DCD line control? */
#undef MODEM_WATCH_CARRIER

/* Do you want hangup to watch for modem DCD line? */
#undef MODEM_HANGUP_WATCH_CARRIER

/* Version string */
#undef RELEASE_VERSION

/* Disable passwords logging? */
#undef BFORCE_LOG_PASSWD

/* Path to the UUCP lock files directory */
#undef BFORCE_LOCK_DIR

/* Do you want use .csy locks? */
#undef BFORCE_USE_CSY

